from __future__ import annotations

import os
from textwrap import dedent
from typing import Any, Callable, TypeVar

FuncType = Callable[..., Any]
F = TypeVar("F", bound=FuncType)

try:
    # Default is DOC_SUB is True
    DOC_SUB = os.getenv("ANALPHIPY_DOC_SUB", "True").lower() not in ("0", "f", "false")
except KeyError:
    DOC_SUB = True


# doc comes from pandas library.  placed directly here because
# 1. This is in a private module of pandas
# 2. pandas not a required dependency of this library.
#
# See https://github.com/pandas-dev/pandas/blob/main/pandas/util/_decorators.py#:~:text=def%20doc


def doc(*docstrings: None | str | Callable, **params) -> Callable[[F], F]:
    """
    A decorator take docstring templates, concatenate them and perform string
    substitution on it.
    This decorator will add a variable "_docstring_components" to the wrapped
    callable to keep track the original docstring template for potential usage.
    If it should be consider as a template, it will be saved as a string.
    Otherwise, it will be saved as callable, and later user __doc__ and dedent
    to get docstring.
    Parameters
    ----------
    *docstrings : None, str, or callable
        The string / docstring / docstring template to be appended in order
        after default docstring under callable.
    **params
        The string which would be used to format docstring template.
    """

    def decorator(decorated: F) -> F:
        # collecting docstring and docstring templates
        docstring_components: list[str | Callable] = []
        if decorated.__doc__:
            docstring_components.append(dedent(decorated.__doc__))

        for docstring in docstrings:
            if docstring is None:
                continue
            if hasattr(docstring, "_docstring_components"):
                # error: Item "str" of "Union[str, Callable[..., Any]]" has no attribute
                # "_docstring_components"
                # error: Item "function" of "Union[str, Callable[..., Any]]" has no
                # attribute "_docstring_components"
                docstring_components.extend(
                    docstring._docstring_components  # type: ignore[union-attr]
                )
            elif isinstance(docstring, str) or docstring.__doc__:
                docstring_components.append(docstring)

        params_applied = [
            component.format(**params)
            if isinstance(component, str) and len(params) > 0
            else component
            for component in docstring_components
        ]

        decorated.__doc__ = "".join(
            [
                component
                if isinstance(component, str)
                else dedent(component.__doc__ or "")
                for component in params_applied
            ]
        )

        # error: "F" has no attribute "_docstring_components"
        decorated._docstring_components = (  # type: ignore[attr-defined]
            docstring_components
        )
        return decorated

    return decorator


def docfiller(*args, **kwargs):
    """To fill common docs.

    Taken from pandas.utils._decorators
    """
    if DOC_SUB:
        return doc(*args, **kwargs)
    else:

        def decorated(func):
            return func

        return decorated


# fmt: off
_shared_docs = {
    "segments":
    """
    segments : sequence of ints
        Integration limits. For ``n = len(segmetns)`` integration will be performed over ranges
        ``(segments[0], segments[1]), (segments[1], segments[2]), ..., (segments[n-2], segments[n-]])``
    """,
    "phi_rep":
    """
    phi_rep : callable
        Repulsive part of pair potential.
    """,
    "beta":
    """
    beta : float
        Inverse temperature.
    """,
    "phi":
    """
    phi : callable
        Potential function.
    """,
    "r":
    """
    r : float or array-like
        Pair separation distance(s).
    """,
    "quad_kws":
    """
    quad_kws : mapping, optional
        Extra arguments to :func:`analphipy.utils.quad_segments`
    """,
    "r_min_exact":
    """
    r_min : float
        Location of minimum in potential energy.
    """,
    "phi_min_exact":
    """
    phi_min : float, optional
        Value of potential energy at minimum.
    """,
    "full_output":
    """
    full_output : bool, optional
        If True, return extra information.
    """,
    "err":
    """
    err : bool, optional
        If True, return error.
    """,
    "error_summed":
    """
    error : float, optional
        Total integration error. Returned if ``err`` or ``full_output`` are `True`.
    """,
    "full_output_summed":
    """
    outputs :
        Output(s) from :func:`scipy.integrate.quad`.  Returned if ``full_output`` is True.
    """,
    "volume_int_func":
    """
    volume : str or callable, optional
        Volume element in integration.
        For example, use ``volume = lambda x: 4 * np.pi * x ** 2`` for spherically symmetric 3d integration.
        Can also pass string value of {'1d', '2d', '3d'}.  If passed None, then assume '1d' integration.
    """,
    # References
    "ref_Noro_Frenkel":
    """
    M.G. Noro and D. Frenkel (2000), "Extended corresponding-states behavior for particles with variable range attractions". Journal of Chemical Physics, 113, 2941.
    """,
    "ref_Barker_Henderson":
    """
    J.A. Barker and D. Henderson (1976), "What Is Liquid? Understanding the States of Matter". Reviews of Modern Physics, 48, 587-671
    """,
    "ref_WCA":
    """
    J.D. Weeks, D. Chandler and H.C. Andersen (1971), "Role of Repulsive Forces in Determining the Equilibrium Structure of Simple Liquids", Journal of Chemical Physics 54, 5237-5247
    """,
    "kl_link":
    """
    `See here for more info <https://en.wikipedia.org/wiki/Kullback%E2%80%93Leibler_divergence#Definition>`_
    """,
    "js_link":
    """
    `See here for more info <https://en.wikipedia.org/wiki/Kullback%E2%80%93Leibler_divergence#Symmetrised_divergence>`_
    """,


}
# fmt: on


def prepare_shared_docs(shared_docs):
    return {k: dedent(v).strip() for k, v in _shared_docs.items()}


_shared_docs = prepare_shared_docs(_shared_docs)

docfiller_shared = docfiller(**_shared_docs)


def factory_docfiller_shared(**kws):

    if len(kws) == 0:
        return docfiller_shared

    else:
        kws = dict(_shared_docs, **prepare_shared_docs(kws))
        return docfiller(**kws)



from __future__ import annotations

from functools import wraps
from typing import Any, Callable, Mapping, Optional

import numpy as np
from scipy.integrate import quad
from scipy.optimize import minimize

from ._docstrings import docfiller_shared
from ._typing import ArrayLike, Float_or_ArrayLike

# from typing_extensions import Protocol

# ArrayLike = Union[Sequence[float], NDArray[np.float_]]
# Float_or_ArrayLike = Union[float, ArrayLike]
# Float_or_Array = Union[float, NDArray[np.float_]]

# Phi_Signature = Callable[..., Union[float, np.ndarray]]

TWO_PI = 2.0 * np.pi

# class Phi_Signature(Protocol):
#     def __call__(self, Float_or_ArrayLike) -> Union[float, np.ndarray]: ...

# Vector = list[float]

# def scale(scalar: float, vector: Vector) -> Vector:
#     return [scalar * num for num in vector]


def combine_segmets(a: ArrayLike, b: ArrayLike) -> list[float]:
    """Combine two lists of segments.

    Parameters
    ----------
    a, b : array-like
        list of segments.

    Returns
    -------
    combined : list of floats
        sorted unique values for ``a`` and ``b``.
    """
    aa, bb = set(a), set(b)
    return sorted(aa.union(bb))


@docfiller_shared
def quad_segments(
    func: Callable,
    segments: ArrayLike,
    args: tuple = (),
    full_output: bool = False,
    sum_integrals: bool = True,
    sum_errors: bool = False,
    err: bool = True,
    **kws,
):
    """
    Perform quadrature with discontinuities

    Parameters
    ----------
    func : callable
        function to be integrated
    {segments}
    args : tuple, optional
        Extra positional arguments to `func`.
    full_output : bool, default=False.
        If True, return extra information.
    sum_integrals : bool, default=True
        If True, sum the segments in the output.
    sum_errors : bool, default=True
        If True and returning `error` sum errors.
    err : bool, default = True
        If True, return error.
    **kws :
        Extra arguments to :func:`scipy.integrate.quad`

    Returns
    -------
    integral : float or list of floats
        If `sum_integrals`,  this is the sum of integrals over each segment.  Otherwise return list
        of values corresponding to integral in each segment.
    errors, float or list of floats, optional
        If `err` or `full_output` are True, then return error.  If `sum_errors`, then sum of errors
        Across segments.
    outputs :
        Output from :func:`scipy.integrate.quad`. If multiple segments, return a list of output.

    See Also
    --------
    scipy.integrate.quad

    """

    out = [
        quad(func, a=a, b=b, args=args, full_output=full_output, **kws)
        for a, b in zip(segments[:-1], segments[1:])
    ]

    if len(segments) == 2:
        if full_output:
            integrals, errors, outputs = out[0]
        else:
            integrals, errors = out[0]
            outputs = None

    else:
        integrals = []
        errors = []

        if full_output:
            outputs = []
            for y, e, o in out:
                integrals.append(y)
                errors.append(e)
                outputs.append(o)
        else:
            outputs = None
            for y, e in out:
                integrals.append(y)
                errors.append(e)

        if sum_integrals:
            integrals = np.sum(integrals)

        if sum_errors:
            errors = np.sum(errors)

    # Gather final results
    result = [integrals]
    if err:
        result.append(errors)
    if outputs is not None:
        result.append(outputs)

    if len(result) == 1:
        return result[0]
    else:
        return result


def minimize_phi(
    phi: Callable, r0: float, bounds: Optional[ArrayLike] = None, **kws
) -> tuple[float, float, Any]:
    """
    Find value of ``r`` which minimized ``phi``.

    Parameters
    ----------
    phi : callable
        Function to be minimized.
    r0 : float
        Guess for position of minimum.
    bounds : tuple of floats
        If passed, should be of form ``bounds=(lower_bound, upper_bound)``.
    **kws :
        Extra arguments to :func:`scipy.optimize.minimize`

    Returns
    -------
    rmin : float
        ``phi(rmin)`` is found location of minimum.
    phimin : float
        Value of ``phi(rmin)``, i.e., the value of ``phi`` at the minimum
    output :
        Output class from :func:`scipy.optimize.minimize`.


    See Also
    --------
    scipy.optimize.minimize
    """

    if bounds is None:
        bounds = (0.0, np.inf)

    if bounds[0] == bounds[1]:
        # force minima to be exactly here
        xmin = bounds[0]
        ymin = phi(xmin)
        outputs = None
    else:

        outputs = minimize(phi, r0, bounds=[bounds], **kws)

        if not outputs.success:
            raise ValueError("could not find min of phi")

        xmin = outputs.x[0]

        try:
            ymin = outputs.fun[0]
        except Exception:
            ymin = outputs.fun

    return xmin, ymin, outputs


# * Phi utilities


def partial_phi(phi, **params):
    from functools import partial

    return partial(phi, **params)


def segments_to_segments_cut(segments, rcut):
    """
    Update segments for 'cut' potential
    """
    return [x for x in segments if x < rcut] + [rcut]


def phi_to_phi_cut(
    rcut: float,
    phi: Callable,
    dphidr: Callable | None = None,
    meta: Mapping | None = None,
):
    r"""
    Create callable ``phi`` and ``dphidr`` cut potentials.

    Parameters
    ----------
    rcut : float
        Postition of cut.
    phi : callable
        input callable function.
    dphidr : callable, optional
    input callable for :math:`d\phi/dr`

    Returns
    -------
    phi_cut : callable
    dphidr_cut : callable or None
        Note that ``dphidr_cut`` is always returned, even if it is None.
    """

    vcut = phi(rcut)

    def phi_cut(r):
        r = np.asarray(r)
        v = np.empty_like(r)

        left = r <= rcut
        right = ~left

        v[right] = 0.0

        if np.any(left):
            v[left] = phi(r[left]) - vcut
        return v

    if dphidr:

        def dphidr_cut(r):

            r = np.asarray(r)
            dvdbeta = np.empty_like(r)

            left = r <= rcut
            right = ~left

            dvdbeta[right] = 0

            if np.any(left):
                dvdbeta[left] = dphidr(r[left])

            return dvdbeta

    else:
        dphidr_cut = None  # type: ignore

    if meta is not None:
        meta = dict(meta, rcut=rcut)
        meta["style"] = meta.get("style", ()) + ("cut",)

    return phi_cut, dphidr_cut, meta


def phi_to_phi_lfs(
    rcut: float,
    phi: Callable,
    dphidr: Callable,
    meta: Mapping | None = None,
):
    r"""
    Create callable ``phi`` and ``dphidr`` cut potentials.

    Parameters
    ----------
    rcut : float
        Postition of cut.
    phi : callable
        input callable function.
    dphidr : callable, optional
    input callable for :math:`d\phi/dr`

    Returns
    -------
    phi_cut : callable
    dphidr_cut : callable or None
        Note that ``dphidr_cut`` is always returned, even if it is None.
    """

    vcut = phi(rcut)
    dvdrcut = dphidr(rcut)

    def phi_cut(r):
        r = np.asarray(r)
        v = np.empty_like(r)

        left = r <= rcut
        right = ~left

        v[right] = 0.0

        if np.any(left):
            v[left] = phi(r[left]) - vcut - dvdrcut * (r - rcut)
        return v

    def dphidr_cut(r):

        r = np.asarray(r)
        dvdbeta = np.empty_like(r)

        left = r <= rcut
        right = ~left

        dvdbeta[right] = 0

        if np.any(left):
            dvdbeta[left] = dphidr(r[left]) - dvdrcut

        return dvdbeta

    if meta is not None:
        meta = dict(meta, rcut=rcut)
        meta["style"] = meta.get("style", ()) + ("lfs",)

    return phi_cut, dphidr_cut, meta


def wca_decomp_rep(phi, dphidr, r_min, phi_min, meta):
    def phi_rep(r: Float_or_ArrayLike) -> np.ndarray:
        """
        WCA repulsive potential
        """
        r = np.array(r)
        v = np.empty_like(r)

        left = r <= r_min
        right = ~left

        v[left] = phi(r[left]) - phi_min
        v[right] = 0.0
        return v

    if dphidr is not None:

        def dphidr_rep(r: Float_or_ArrayLike) -> np.ndarray:

            r = np.array(r)
            dvdr = np.empty_like(r)

            left = r <= r_min
            right = ~left

            dvdr[left] = dphidr(r[left])
            dvdr[right] = 0.0

            return dvdr

    else:
        dphidr_rep = None

    if meta is not None:
        meta = dict(meta)
        meta["style"] = meta.get("style", ()) + ("wca_rep",)

    return phi_rep, dphidr_rep, meta


def wca_decomp_att(phi, dphidr, r_min, phi_min, meta):
    def phi_att(r: Float_or_ArrayLike) -> np.ndarray:
        """
        WCA repulsive potential
        """
        r = np.array(r)
        v = np.empty_like(r)

        left = r <= r_min
        right = ~left

        v[left] = phi_min
        v[right] = phi(r[right])
        return v

    if dphidr is not None:

        def dphidr_att(r: Float_or_ArrayLike) -> np.ndarray:

            r = np.array(r)
            dvdr = np.empty_like(r)

            left = r <= r_min
            right = ~left

            dvdr[left] = 0.0
            dvdr[right] = dphidr(r[right])

            return dvdr

    else:
        dphidr_att = None

    if meta is not None:
        meta = dict(meta)
        meta["style"] = meta.get("style", ()) + ("wca_att",)

    return phi_att, dphidr_att, meta


def add_quad_kws(func: Callable) -> Callable:
    @wraps(func)
    def wrapped(self, *args, **kws):
        kws = dict(self.quad_kws, **kws)
        return func(self, *args, **kws)

    return wrapped



#!/usr/bin/env python
#
# python_boilerplate documentation build configuration file, created by
# sphinx-quickstart on Fri Jun  9 13:47:02 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another
# directory, add these directories to sys.path here. If the directory is
# relative to the documentation root, use os.path.abspath to make it
# absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath(".."))

import analphipy

# -- General configuration ---------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    # "sphinx.ext.extlinks",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "IPython.sphinxext.ipython_directive",
    "IPython.sphinxext.ipython_console_highlighting",
    "nbsphinx",
    # "sphinx_autosummary_accessors",
    # "scanpydoc.rtd_github_links",
    # local source
    # "sphinx.ext.viewcode",
    # github source
    "sphinx.ext.linkcode",
    # spelling
    "sphinxcontrib.spelling",
]

# defined stuff, from xarray
nbsphinx_prolog = """

{% set docname = env.doc2path(env.docname, base=None) %}


You can view this notebook `on Github <https://github.com/usnistgov/analphipy/blob/main/doc/{{ docname }}>`_.
"""

nbsphinx_kernel_name = "python3"
autosummary_generate = True
# autoclass_content = "both"  # include both class docstring and __init__
autodoc_default_flags = [
    # Make sure that any autodoc declarations show the right members
    "members",
    "inherited-members",
    "private-members",
    "show-inheritance",
]
# # for scanpydoc's jinja filter
# project_dir = pathlib.Path(__file__).parent.parent
html_context = {
    "github_user": "usnistgov",
    "github_repo": "analphipy",
    "github_version": "main",
    "doc_path": "doc",
}

autodoc_typehints = "none"

napoleon_google_docstring = False
napoleon_numpy_docstring = True

napoleon_use_param = False
napoleon_use_rtype = True
napoleon_preprocess_types = True
napoleon_type_aliases = {
    # general terms
    "sequence": ":term:`sequence`",
    "iterable": ":term:`iterable`",
    "callable": ":py:func:`callable`",
    "dict_like": ":term:`dict-like <mapping>`",
    "dict-like": ":term:`dict-like <mapping>`",
    "path-like": ":term:`path-like <path-like object>`",
    "mapping": ":term:`mapping`",
    "hashable": ":term:`hashable`",
    # special terms
    # "same type as caller": "*same type as caller*",  # does not work, yet
    # "same type as values": "*same type as values*",  # does not work, yet
    # stdlib type aliases
    "MutableMapping": "~collections.abc.MutableMapping",
    "sys.stdout": ":obj:`sys.stdout`",
    "timedelta": "~datetime.timedelta",
    "string": ":class:`string <str>`",
    # numpy terms
    "array_like": ":term:`array_like`",
    "array-like": ":term:`array-like <array_like>`",
    "scalar": ":term:`scalar`",
    "array": ":term:`array`",
    # matplotlib terms
    "color-like": ":py:func:`color-like <matplotlib.colors.is_color_like>`",
    "matplotlib colormap name": ":doc:`matplotlib colormap name <matplotlib:gallery/color/colormap_reference>`",
    "matplotlib axes object": ":py:class:`matplotlib axes object <matplotlib.axes.Axes>`",
    "colormap": ":py:class:`colormap <matplotlib.colors.Colormap>`",
    # objects without namespace: xarray
    "DataArray": "~xarray.DataArray",
    "Dataset": "~xarray.Dataset",
    # objects without namespace: numpy
    "ndarray": "~numpy.ndarray",
    "MaskedArray": "~numpy.ma.MaskedArray",
    "dtype": "~numpy.dtype",
    "ComplexWarning": "~numpy.ComplexWarning",
    # objects with namespace analphipy
    "lnPiMasked": "~lnpy.lnPiMasked",
    "lnPiCollection": "~lnpy.lnPiCollection",
}


numpydoc_class_members_toctree = True
numpydoc_show_class_members = False


# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "analphipy"
copyright = "2021, William P. Krekelberg"
author = "William P. Krekelberg"

# The version info for the project you're documenting, acts as replacement
# for |version| and |release|, also used in various other places throughout
# the built documents.
#
# The short X.Y version.
version = analphipy.__version__
# The full version, including alpha/beta/rc tags.
release = analphipy.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "**.ipynb_checkpoints"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# based on numpy doc/source/conf.py
def linkcode_resolve(domain, info):
    """
    Determine the URL corresponding to Python object
    """
    import inspect

    if domain != "py":
        return None

    modname = info["module"]
    fullname = info["fullname"]

    submod = sys.modules.get(modname)
    if submod is None:
        return None

    obj = submod
    for part in fullname.split("."):
        try:
            obj = getattr(obj, part)
        except AttributeError:
            return None

    try:
        fn = inspect.getsourcefile(inspect.unwrap(obj))
    except TypeError:
        fn = None
    if not fn:
        return None

    try:
        source, lineno = inspect.getsourcelines(obj)
    except OSError:
        lineno = None

    if lineno:
        linespec = f"#L{lineno}-L{lineno + len(source) - 1}"
    else:
        linespec = ""

    fn = os.path.relpath(fn, start=os.path.dirname(analphipy.__file__))

    return (
        f"https://github.com/usnistgov/analphipy/blob/master/analphipy/{fn}{linespec}"
    )


spelling_word_list_filename = "spelling_wordlist.txt"

# -- Options for HTML output -------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = 'alabaster'
html_theme = "sphinx_rtd_theme"


# Theme options are theme-specific and customize the look and feel of a
# theme further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {"logo_only": True}


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


# Sometimes the savefig directory doesn't exist and needs to be created
# https://github.com/ipython/ipython/issues/8733
# becomes obsolete when we can pin ipython>=5.2; see ci/requirements/doc.yml
ipython_savefig_dir = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "_build", "html", "_static"
)
if not os.path.exists(ipython_savefig_dir):
    os.makedirs(ipython_savefig_dir)


# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
today_fmt = "%Y-%m-%d"
html_last_updated_fmt = today_fmt


# -- Options for HTMLHelp output ---------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "analphipydoc"


# -- Options for LaTeX output ------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "analphipy.tex",
        "analphipy Documentation",
        "William P. Krekelberg",
        "manual",
    ),
]


# -- Options for manual page output ------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "analphipy", "analphipy Documentation", [author], 1)]


# -- Options for Texinfo output ----------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "analphipy",
        "analphipy Documentation",
        author,
        "analphipy",
        "One line description of project.",
        "Miscellaneous",
    ),
]


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    # "scipy": ('https://docs.scipy.org/doc/scipy/reference/', None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
    "numba": ("https://numba.pydata.org/numba-doc/latest", None),
    # "matplotlib": ("https://matplotlib.org", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "dask": ("https://docs.dask.org/en/latest", None),
    "cftime": ("https://unidata.github.io/cftime", None),
    "sparse": ("https://sparse.pydata.org/en/latest/", None),
    "xarray": ("https://docs.xarray.dev/en/stable/", None),
    "skimage": ("https://scikit-image.org/docs/stable", None),
}

# think jinja stuff
# def escape_underscores(string):
#     return string.replace("_", r"\_")


# def setup(app):
#     DEFAULT_FILTERS["escape_underscores"] = escape_underscores



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Mapping, Optional, Sequence, cast

import numpy as np
from custom_inherit import doc_inherit

from ._docstrings import docfiller_shared
from ._typing import ArrayLike, Float_or_Array, Float_or_ArrayLike, Phi_Signature
from .cached_decorators import gcached
from .measures import secondvirial, secondvirial_dbeta, secondvirial_sw
from .utils import TWO_PI, add_quad_kws, minimize_phi, quad_segments

if TYPE_CHECKING:
    from ._potentials import PhiBase  # type: ignore


@docfiller_shared
def sig_nf(
    phi_rep: Phi_Signature,
    beta: float,
    segments: ArrayLike,
    err: bool = False,
    full_output: bool = False,
    **kws,
):
    r"""Noro-Frenkel/Barker-Henderson effective hard sphere diameter.

    This is calculated using the formula [1]_ [2]_

    .. math::

        \sigma_{{\rm BH}}(\beta) = \int_0^{{\infty}} dr \left( 1 - \exp[-\beta \phi_{{\rm rep}}(r)]\right)

    where :math:`\phi_{{\rm rep}}(r)` is the repulsive part of the potential [3]_.

    Parameters
    ----------
    {phi_rep}
    {beta}
    {segments}
    phi_rep : callable
        Repulsive part of pair potential.
    beta : float
        Inverse temperature.
    segments : array-like
        Integration segments.
    err : bool, default=False
        If True, return error value.
    full_output : bool, default=True
        If True, return full_output.

    Returns
    -------
    sig_nf : float or list of floats
        Value of integral.
    errors, float or list of floats, optional
        If `err` or `full_output` are True, then return sum of errors.
    outputs :
        Output from :func:`scipy.integrate.quad`. If multiple segments, return a list of output.



    See Also
    --------
    ~analphipy.utils.quad_segments

    References
    ----------
    .. [1] {ref_Noro_Frenkel}

    .. [2] {ref_Barker_Henderson}

    .. [3] {ref_WCA}

    """

    def integrand(r):
        return 1.0 - np.exp(-beta * phi_rep(r))

    return quad_segments(
        integrand,
        segments=segments,
        sum_integrals=True,
        sum_errors=True,
        err=err,
        full_output=full_output,
        **kws,
    )


@doc_inherit(sig_nf, style="numpy_with_merge")
def sig_nf_dbeta(
    phi_rep: Phi_Signature,
    beta: float,
    segments: ArrayLike,
    err: bool = False,
    full_output: bool = False,
    **kws,
):
    r"""
    Derivative with respect to inverse temperature ``beta`` of ``sig_nf``.


    See refs [1]_ [2]_ [3]_

    .. math::

        \frac{d \sigma_{\rm BH}}{d\beta} = \int_0^{\infty} dr \phi_{\rm rep}(r) \exp[-\beta \phi_{\rm rep}(r)]

    See Also
    --------
    sig_nf

    """

    def integrand(r):
        v = phi_rep(r)
        if np.isinf(v):
            return 0.0
        else:
            return v * np.exp(-beta * v)

    return quad_segments(
        integrand,
        segments=segments,
        sum_integrals=True,
        sum_errors=True,
        err=err,
        full_output=full_output,
        **kws,
    )


@docfiller_shared
def lam_nf(beta: float, sig: float, eps: float, B2: float):
    r"""
    Noro-Frenkel effective lambda parameter

    This is the value of :math:`\lambda` in a square well potential which matches second virial
    coefficients.  The square well fluid is defined as [1]_

    .. math::

        \phi_{{\rm sw}}(r) =
        \begin{{cases}}
            \infty & r \leq \sigma \\
            \epsilon &  \sigma < r \leq \lambda \sigma \\
            0 &  r > \lambda \sigma
        \end{{cases}}


    Parameters
    ----------
    {beta}
    sig : float
        Particle diameter.
    eps : float
        Energy parameter in square well potential. The convention is that ``eps`` is the same as the value of ``phi`` at the minimum.
    B2 : float
        Second virial coefficient to match.

    Returns
    -------
    lam_nf : float
        Value of `lambda` in an effective square well fluid which matches ``B2``.


    References
    ----------
    .. [1] {ref_Noro_Frenkel}
    """
    B2star = B2 / (TWO_PI / 3.0 * sig**3)

    # B2s_SW = 1 + (1-exp(-beta epsilon)) * (lambda**3 - 1)
    #        = B2s
    lam = ((B2star - 1.0) / (1.0 - np.exp(-beta * eps)) + 1.0) ** (1.0 / 3.0)
    return lam


@docfiller_shared
def lam_nf_dbeta(
    beta: float,
    sig: float,
    eps: float,
    lam: float,
    B2: float,
    B2_dbeta: float,
    sig_dbeta: float,
):
    """
    Calculate derivative of ``lam_nf``  with respect to ``beta``

    Parameters
    ----------
    {beta}
    sig : float
        Particle diameter.
    eps : float
        Energy parameter in square well potential. The convention is that ``eps`` is the same as the value of ``phi`` at the minimum.
    B2 : float
        Second virial coefficient to match.
    lam : float
        Value from :func:`analphipy.norofrenekl.lam_nf`.
    B2_dbeta : float
        d(B2)/d(beta) at ``beta``
    sig_dbeta : float
        derivative of Noro-Frenkel sigma w.r.t inverse temperature at `beta`

    Returns
    -------
    lam_nf_dbeta : float
        Value of ``d(lam_nf)/d(beta)``

    See Also
    --------
    lam_nf

    """

    B2_hs = TWO_PI / 3.0 * sig**3

    dB2stardbeta = 1.0 / B2_hs * (B2_dbeta - B2 * 3.0 * sig_dbeta / sig)

    # e = np.exp(-beta * eps)
    # f = 1. - e

    e = np.exp(beta * eps)

    out = 1.0 / (3 * lam**2 * (e - 1)) * (dB2stardbeta * e - eps * (lam**3 - 1))

    return out


@docfiller_shared
class NoroFrenkelPair:
    """
    Class to calculate Noro-Frenkel parameters

    See [1]_ [2]_ [3]_


    Parameters
    ----------
    {phi}
    {segments}
    {r_min_exact}
    {phi_min_exact}
    {quad_kws}

    References
    ----------
    .. [1] {ref_Noro_Frenkel}

    .. [2] {ref_Barker_Henderson}

    .. [3] {ref_WCA}
    """

    def __init__(
        self,
        phi: Phi_Signature,
        segments: ArrayLike,
        r_min: float,
        phi_min: Float_or_Array,
        quad_kws: Optional[Mapping[str, Any]] = None,
    ):

        self.phi = phi
        self.r_min = r_min

        if phi_min is None:
            phi_min = phi(r_min)
        self.phi_min = phi_min
        self.segments = segments

        if quad_kws is None:
            quad_kws = {}
        self.quad_kws = quad_kws

    def __repr__(self):

        params = ",\n    ".join(
            [
                f"{v}={getattr(self, v)}"
                for v in ["phi", "segments", "r_min", "phi_min", "quad_kws"]
            ]
        )

        return f"{type(self).__name__}({params})"

    @docfiller_shared
    def phi_rep(self, r: Float_or_ArrayLike) -> np.ndarray:
        """
        Repulsive part of potential

        This is the Weeks-Chandler-Anderson decomposition.

        Parameters
        ----------
        {r}

        Returns
        -------
        output : float or ndarray
            Value of ``phi_ref`` at separation(s) ``r``.

        """
        r = np.array(r)
        phi = np.empty_like(r)
        m = r <= self.r_min

        phi[m] = self.phi(r[m]) - self.phi_min
        phi[~m] = 0.0

        return phi

    @classmethod
    @docfiller_shared
    def from_phi(
        cls,
        phi: Phi_Signature,
        segments: ArrayLike,
        r_min: Optional[float] = None,
        bounds: Optional[ArrayLike] = None,
        quad_kws: Optional[Mapping[str, Any]] = None,
        **kws,
    ) -> NoroFrenkelPair:
        """
        Create object from pair potential function


        Parameters
        ----------
        {phi}
        {segments}
        r_min : float, optional
            Optional guess for numerically finding minimum in `phi`.
        bounds : array-like, optional
            Optional bounds for numerically locating ``r_min``.
        quad_kws : mapping, optional
            Optional arguments to :func:`analphipy.utils.quad_segments`.
        **kws :
            Extra arguments to :func:`analphipy.utils.minimize_phi`.

        Returns
        -------
        output : instance of calling class
        """

        if bounds is None:
            bounds = (segments[0], segments[-1])

        if bounds[-1] == np.inf and r_min is None:
            raise ValueError("if specify infinite bounds, must supply guess")

        if r_min is None:
            r_min = cast(float, np.mean(bounds))

        r_min, phi_min, _ = minimize_phi(
            phi, r0=cast(float, r_min), bounds=bounds, **kws
        )
        return cls(
            phi=phi, r_min=r_min, phi_min=phi_min, segments=segments, quad_kws=quad_kws
        )

    @classmethod
    def from_phi_class(
        cls,
        phi: "PhiBase",
        r_min: Optional[float] = None,
        bounds: Optional[tuple[float, float]] = None,
        quad_kws: Optional[dict[str, Any]] = None,
        **kws,
    ):
        """
        Create object, trying to use pre computed values for ``r_min``, ``phi_min``.

        Parameters
        ----------
        phi : Phi_Abstracclass instance
        r_min : float, optional
            Optional guess for numerically finding minimum in `phi`.
        bounds : array-like, optional
            Optional bounds for numerically locating ``r_min``.
        quad_kws : mapping, optional
            Optional arguments to :func:`analphipy.utils.quad_segments`.
        **kws :
            Extra arguments to :func:`analphipy.utils.minimize_phi`.


        Returns
        -------
        output : instance of calling class
        """

        try:
            return cls(
                phi=phi.phi,
                segments=phi.segments,
                r_min=phi.r_min,
                phi_min=phi.phi_min,
                quad_kws=quad_kws,
            )
        except NotImplementedError:
            return cls.from_phi(
                phi=phi.phi,
                segments=phi.segments,
                r_min=r_min,
                bounds=bounds,
                quad_kws=quad_kws,
                **kws,
            )

    @gcached(prop=False)
    @add_quad_kws
    def secondvirial(self, beta: float, **kws):
        """Second virial coefficient.

        See Also
        --------
        ~measures.secondvirial
        """
        return secondvirial(phi=self.phi, beta=beta, segments=self.segments, **kws)

    @gcached()
    def _segments_rep(self) -> list:
        return [x for x in self.segments if x < self.r_min] + [self.r_min]

    @gcached(prop=False)
    @add_quad_kws
    def sig(self, beta: float, **kws):
        """Effective hard sphere diameter.

        See Also
        --------
        ~analphipy.norofrenkel.sig_nf

        """
        return sig_nf(self.phi_rep, beta=beta, segments=self._segments_rep, **kws)

    def eps(self, beta: float, **kws) -> float:
        """Effective square well epsilon.

        This is equal to the minimum in ``phi``.
        """
        return cast(float, self.phi_min)

    @gcached(prop=False)
    @add_quad_kws
    def lam(self, beta: float, **kws):
        """Effective square well lambda.

        See Also
        --------
        ~norofrenkel.lam_nf
        """
        return lam_nf(
            beta=beta,
            sig=self.sig(beta, **kws),
            eps=self.eps(beta, **kws),
            B2=self.secondvirial(beta, **kws),
        )

    @gcached(prop=False)
    @add_quad_kws
    def sw_dict(self, beta: float, **kws) -> dict[str, float]:
        """Dictionary view of Noro-Frenkel parameters."""
        sig = self.sig(beta, **kws)
        eps = self.eps(beta, **kws)
        lam = lam_nf(beta=beta, sig=sig, eps=eps, B2=self.secondvirial(beta, **kws))
        return {"sig": sig, "eps": eps, "lam": lam}

    @gcached(prop=False)
    @add_quad_kws
    def secondvirial_dbeta(self, beta: float, **kws):
        """Derivative of ``secondvirial`` with respect to ``beta``

        See Also
        --------
        ~measures.secondvirial_dbeta
        """
        return secondvirial_dbeta(
            phi=self.phi, beta=beta, segments=self.segments, **kws
        )

    @gcached(prop=False)
    @add_quad_kws
    def sig_dbeta(self, beta: float, **kws):
        """Derivative of effective hard-sphere diameter with respect to ``beta``

        See Also
        --------
        ~norofrenkel.sig_nf_dbeta

        """
        return sig_nf_dbeta(self.phi_rep, beta=beta, segments=self.segments, **kws)

    @gcached(prop=False)
    def lam_dbeta(self, beta: float, **kws):
        """Derivative of effective lambda parameter with respect to ``beta``.

        See Also
        --------
        ~norofrenkel.lam_nf_dbeta
        """
        return lam_nf_dbeta(
            beta=beta,
            sig=self.sig(beta, **kws),
            eps=self.eps(beta, **kws),
            lam=self.lam(beta, **kws),
            B2=self.secondvirial(beta, **kws),
            B2_dbeta=self.secondvirial_dbeta(beta, **kws),
            sig_dbeta=self.sig_dbeta(beta, **kws),
        )

    @gcached(prop=False)
    def secondvirial_sw(self, beta: float, **kws):
        """Second virial coefficient of effective square well fluid.

        For testing.  This should be the same of value from :meth:`secondvirial`
        """
        return secondvirial_sw(
            beta=beta,
            sig=self.sig(beta, **kws),
            eps=self.eps(beta, **kws),
            lam=self.lam(beta, **kws),
        )

    def B2(self, beta: float, **kws):
        """Alias to :meth:`secondvirial`"""
        return self.secondvirial(beta, **kws)

    def B2_dbeta(self, beta: float, **kws):
        """Alias to :meth:`secondvirial_dbeta`"""
        return self.secondvirial_dbeta(beta, **kws)

    def B2_sw(self, beta, **kws):
        """Alias to :meth:`secondvirial_sw`"""
        return self.secondvirial_sw(beta, **kws)

    def table(
        self,
        betas: ArrayLike,
        props: Sequence[str] = None,
        key_format: str = "{prop}",
        **kws,
    ) -> dict[str, Any]:
        """
        Create a dictionary of outputs for multiple values of inverse temperature ``beta``.

        Parameters
        ----------
        betas : array-like
            Array of values of inverse temperature ``beta``.
        props : sequence of strings
            Name of methods to access.
        key_format : string, default="{prop}"
            Each key in the output dictionary will have the value ``key_format.format(prop=prop)``

        Returns
        -------
        output : dict
            dictionary of arrays.
        """
        if props is None:
            props = ["B2", "sig", "eps", "lam"]

        table = {"beta": betas}

        for prop in props:
            f = getattr(self, prop)
            key = key_format.format(prop=prop)
            table[key] = [f(beta=beta, **kws) for beta in betas]
        return table



from __future__ import annotations

from typing import Callable, Sequence, Union

import numpy as np
from numpy.typing import NDArray  # , ArrayLike

# if TYPE_CHECKING:
#     from .potentials_generic import PhiBase


# T_PhiBase = TypeVar("T_PhiBase", bound="PhiBase")


ArrayLike = Union[Sequence[float], NDArray[np.float_]]
Float_or_ArrayLike = Union[float, ArrayLike]
Float_or_Array = Union[float, NDArray[np.float_]]
Phi_Signature = Callable[..., Union[float, np.ndarray]]



from __future__ import annotations

from typing import Callable, Mapping, Optional, Sequence, Union, cast

import numpy as np
from custom_inherit import doc_inherit

from analphipy.cached_decorators import gcached

from ._docstrings import docfiller_shared
from ._typing import ArrayLike, Float_or_ArrayLike, Phi_Signature
from .utils import TWO_PI, add_quad_kws, combine_segmets, quad_segments


def other():
    """A function"""
    pass


@docfiller_shared
def secondvirial(
    phi: Phi_Signature,
    beta: float,
    segments: ArrayLike,
    err: bool = False,
    full_output: bool = False,
    **kws,
):
    r"""
    Calculate the second virial coefficient.

    .. math::

        B_2(\beta) = -\int 2\pi r^2 dr \left(\exp(-\beta \phi(r)) - 1\right)

    Parameters
    ----------
    {phi}
    {beta}
    {segments}
    {err}
    {full_output}
    **kws
        Extra arguments to :func:`analphipy.quad_segments`

    Returns
    -------
    B2 : float
        Value of second virial coefficient.
    {error_summed}
    {full_output_summed}

    See Also
    --------
    ~analphipy.utils.quad_segments
    """

    def integrand(r):
        return TWO_PI * r**2 * (1 - np.exp(-beta * phi(r)))

    return quad_segments(
        integrand,
        segments=segments,
        sum_integrals=True,
        sum_errors=True,
        err=err,
        full_output=full_output,
        **kws,
    )


@docfiller_shared
def secondvirial_dbeta(
    phi: Phi_Signature,
    beta: float,
    segments: ArrayLike,
    err: bool = False,
    full_output: bool = False,
    **kws,
):
    r"""
    ``beta`` derivative of second virial coefficient.

    .. math::

        \frac{{d B_2}}{{d \beta}} = \int 2\pi r^2 dr \phi(r) \exp(-\beta \phi(r))

    Parameters
    ----------
    {phi}
    {beta}
    {segments}
    {err}
    {full_output}

    Returns
    -------
    dB2dbeta : float
        Value of derivative.
    {error_summed}
    {full_output_summed}


    """

    def integrand(r):
        v = phi(r)
        if np.isinf(v):
            return 0.0
        else:
            return TWO_PI * r**2 * v * np.exp(-beta * v)

    return quad_segments(
        integrand,
        segments=segments,
        sum_integrals=True,
        sum_errors=True,
        err=err,
        full_output=full_output,
        **kws,
    )


@docfiller_shared
def secondvirial_sw(beta: float, sig: float, eps: float, lam: float):
    r"""
    Second virial coefficient for a square well (SW) fluid. Note that this assumes that
    the SW fluid is defined by the potential:

    .. math::

        \phi(r) =
        \begin{{cases}}
            \infty & r \leq \sigma \\
            \epsilon & \sigma < r \leq \lambda \sigma  \\
            0 & \lambda \sigma < r
        \end{{cases}}


    Parameters
    ----------
    {beta}
    sig : float
        Length parameter :math:`\sigma`.
    eps : float
        Energy parameter :math:`\epsilon`.
    lam : float
        Well width parameter :math:`\lambda`.

    Returns
    -------
    B2 : float
        Value of second virial coefficient.

    """
    return (
        TWO_PI / 3.0 * sig**3 * (1.0 + (1 - np.exp(-beta * eps)) * (lam**3 - 1.0))
    )


def diverg_kl_integrand(
    p: Float_or_ArrayLike,
    q: Float_or_ArrayLike,
    volume: Optional[Float_or_ArrayLike] = None,
) -> np.ndarray:
    p, q = np.asarray(p), np.asarray(q)

    out = np.empty_like(p)

    zero = p == 0.0
    hero = ~zero

    out[zero] = 0.0
    out[hero] = p[hero] * np.log(p[hero] / q[hero])

    if volume is not None:
        out *= np.asarray(volume)

    return out


@docfiller_shared
def diverg_kl_disc(
    P: Float_or_ArrayLike, Q: Float_or_ArrayLike, axis: Optional[int] = None
) -> Union[float, np.ndarray]:
    """
    Calculate discrete Kullback–Leibler divergence

    Parameteres
    -----------
    P, Q : array-like
        Probabilities to consider
    Returns
    -------
    result : float or ndarray
        value of KB divergence

    References
    ----------
    {kl_link}
    """

    P, Q = np.asarray(P), np.asarray(Q)
    out = diverg_kl_integrand(P, Q).sum(axis=axis)

    return cast(Union[float, np.ndarray], out)


def _check_volume_func(volume: Optional[Union[str, Callable]] = None) -> Callable:
    if volume is None:
        volume = "1d"

    if isinstance(volume, str):
        if volume == "1d":
            volume = lambda x: 1.0
        elif volume == "2d":
            volume = lambda x: 2.0 * np.pi * x
        elif volume == "3d":
            volume = lambda x: 4 * np.pi * x**2
        else:
            raise ValueError("unknown dimension")
    else:
        assert callable(volume)

    return volume


@docfiller_shared
def diverg_kl_cont(
    p: Callable,
    q: Callable,
    segments: ArrayLike,
    segments_q: Optional[ArrayLike] = None,
    volume: Optional[Union[str, Callable]] = None,
    err: bool = False,
    full_output: bool = False,
    **kws,
):
    """
    Calculate continuous Kullback–Leibler divergence for continuous pdf

    Parameters
    ----------
    p, q: callable
        Probabilities to consider
    {volume_int_func}
    {segments}
    segments_q : list, optional
        if supplied, build total segments by combining segments and segments_q
    {err}
    {full_output}

    Returns
    -------
    result : float or ndarray
        value of KB divergence
    {error_summed}
    {full_output_summed}

    References
    ----------
    {kl_link}
    """
    volume = _check_volume_func(volume)

    if segments_q is not None:
        segments = combine_segmets(segments, segments_q)

    def func(x):
        return diverg_kl_integrand(p=p(x), q=q(x), volume=volume(x))

    return quad_segments(
        func,
        segments=segments,
        sum_integrals=True,
        sum_errors=True,
        err=err,
        full_output=full_output,
        **kws,
    )


@doc_inherit(diverg_kl_disc, style="numpy_with_merge")
def diverg_js_disc(
    P: Float_or_ArrayLike, Q: Float_or_ArrayLike, axis: Optional[int] = None
) -> Union[float, np.ndarray]:
    """
    Discrete Jensen–Shannon divergence

    Returns
    -------
    result : float or ndarray
        value of JS divergence

    References
    ----------
    {js_link}
    """

    P, Q = np.asarray(P), np.asarray(Q)

    M = 0.5 * (P + Q)

    out = 0.5 * (diverg_kl_disc(P, M, axis=axis) + diverg_kl_disc(Q, M, axis=axis))
    return cast(Union[float, np.ndarray], out)


def diverg_js_integrand(
    p: Float_or_ArrayLike,
    q: Float_or_ArrayLike,
    volume: Optional[Float_or_ArrayLike] = None,
) -> np.ndarray:
    p = np.asarray(p)
    q = np.asarray(q)

    m = 0.5 * (p + q)

    out = 0.5 * (diverg_kl_integrand(p, m) + diverg_kl_integrand(q, m))

    if volume is not None:
        out *= np.asarray(volume)

    return cast(np.ndarray, out)


@doc_inherit(diverg_kl_cont, style="numpy_with_merge")
@docfiller_shared
def diverg_js_cont(
    p: Callable,
    q: Callable,
    segments: ArrayLike,
    segments_q: Optional[ArrayLike] = None,
    volume: Optional[Union[str, Callable]] = None,
    err: bool = False,
    full_output: bool = False,
    **kws,
):
    """
    Continuous Jensen–Shannon divergence


    Returns
    -------
    result : float or ndarray
        value of JS divergence

    References
    ----------
    {js_link}
    """

    volume = _check_volume_func(volume)

    if segments_q is not None:
        segments = combine_segmets(segments, segments_q)

    def func(x):
        return diverg_js_integrand(p(x), q(x), volume(x))

    return quad_segments(
        func,
        segments=segments,
        sum_integrals=True,
        sum_errors=True,
        err=err,
        full_output=full_output,
        **kws,
    )


@docfiller_shared
class Measures:
    """
    Convenience class for calculating measures

    Parameters
    ----------
    {phi}
    {segments}
    {quad_kws}

    """

    def __init__(
        self,
        phi: Phi_Signature,
        segments: Sequence[float],
        quad_kws: Mapping | None = None,
    ) -> None:

        self.phi = phi
        self.segments = segments
        if quad_kws is None:
            quad_kws = {}
        self.quad_kws = quad_kws

    @gcached(prop=False)
    @add_quad_kws
    @docfiller_shared
    def secondvirial(self, beta, err=False, full_output=False, **kws):
        """
        Calculate second virial coefficient.

        Parameters
        ----------
        {beta}
        {err}
        {full_output}

        **kws
            Extra arguments to :func:`analphipy.quad_segments`

        Returns
        -------
        B2 : float
            Value of second virial coefficient.
        {error_summed}
        {full_output_summed}

        See Also
        --------
        ~measures.secondvirial
        """
        return secondvirial(
            phi=self.phi,
            beta=beta,
            segments=self.segments,
            err=err,
            full_output=full_output,
            **kws,
        )

    @gcached(prop=False)
    @add_quad_kws
    @docfiller_shared
    def secondvirial_dbeta(self, beta, err=False, full_output=False, **kws):
        """
        Calculate ``beta`` derivative of second virial coefficient.

        Parameters
        ----------
        {beta}
        {err}
        {full_output}

        Returns
        -------
        dB2dbeta : float
            Value of derivative.
        {error_summed}
        {full_output_summed}


        See Also
        --------
        ~measures.secondvirial_dbeta
        """
        return secondvirial_dbeta(
            phi=self.phi,
            beta=beta,
            segments=self.segments,
            err=err,
            full_output=full_output,
            **kws,
        )

    @docfiller_shared
    @add_quad_kws
    def boltz_diverg_js(
        self,
        other,
        beta: float,
        beta_other: float | None = None,
        volume: str | Callable = "3d",
        err: bool = False,
        full_output: bool = False,
        **kws,
    ):
        r"""
        Jensen-Shannon divergence of the Boltzmann factors of two potentials.

        The Boltzmann factors are defined as:

        .. math::

            B(r; \beta, \phi) = \exp(-\beta \phi(r))


        Parameters
        ----------
        other : :class:`analphipy.PhiBase`
            Class wrapping other potential to compare `self` to.
        {beta}
        beta_other : float, optional
            beta value to evaluate other Boltzmann factor at.
        {volume_int_func}

        See Also
        --------
        ~measures.diverg_js_cont

        References
        --------
        `See here for more info <https://en.wikipedia.org/wiki/Kullback%E2%80%93Leibler_divergence#Symmetrised_divergence>`
        """

        if beta_other is None:
            beta_other = beta

        p = lambda x: np.exp(-beta * self.phi(x))
        q = lambda x: np.exp(-beta_other * other.phi(x))

        return diverg_js_cont(
            p=p,
            q=q,
            segments=self.segments,
            segments_q=other.segments,
            volume=volume,
            err=err,
            full_output=full_output,
            **kws,
        )

    def mayer_diverg_js(
        self,
        other,
        beta: float,
        beta_other: float | None = None,
        volume: str | Callable = "3d",
        err: bool = False,
        full_output: bool = False,
        **kws,
    ):
        r"""
        Jensen-Shannon divergence of the Mayer f-functions of two potentials.

        The Mayer f-function is defined as:

        .. math::

            f(r; \beta, \phi) = \exp(-beta \phi(r)) - 1


        Parameters
        ----------
        other : :class:`analphipy.PhiBase`
            Class wrapping other potential to compare `self` to.
        {beta}
        beta_other : float, optional
            beta value to evaluate other Boltzmann factor at.
        {volume_int_func}


        See Also
        --------
        ~measures.diverg_js_cont


        References
        ----------
        `See here for more info <https://en.wikipedia.org/wiki/Kullback%E2%80%93Leibler_divergence#Symmetrised_divergence>`
        """

        if beta_other is None:
            beta_other = beta

        p = lambda x: np.exp(-beta * self.phi(x)) - 1.0
        q = lambda x: np.exp(-beta_other * other.phi(x)) - 1.0

        return diverg_js_cont(
            p=p,
            q=q,
            segments=self.segments,
            segments_q=other.segments,
            volume=volume,
            err=err,
            full_output=full_output,
            **kws,
        )



import numpy as np

import analphipy.potentials as pots

# import pytest


def _do_test(params, factory, kws=None, cut=False, lfs=False, phidphi=True):

    if kws is None:
        kws = params.asdict()

    if cut or lfs:
        rcut = kws.pop("rcut")
    else:
        rcut = None

    p = factory(**kws)

    if cut:
        p = p.cut(rcut=rcut)
    if lfs:
        p = p.lfs(rcut=rcut)

    if phidphi:
        phi0, dphi0 = params.get_phidphi()

        phi1 = p.phi(params.r)
        dphi1 = -p.dphidr(params.r) / params.r

        np.testing.assert_allclose(dphi0, dphi1)

    else:
        phi0 = params.get_phi()
        phi1 = p.phi(params.r)

    np.testing.assert_allclose(phi0, phi1)


def test_lj(lj_params):
    _do_test(lj_params, pots.Phi_lj)


def test_lj_cut(lj_cut_params):
    _do_test(lj_cut_params, pots.Phi_lj, cut=True)


def test_lj_lfs(lj_lfs_params):
    _do_test(lj_lfs_params, pots.Phi_lj, lfs=True)


def test_nm(nm_params):
    _do_test(nm_params, pots.Phi_nm)


def test_nm_cut(lj_cut_params):
    params = lj_cut_params

    phi0, dphi0 = params.get_phidphi()

    p = pots.Phi_nm(n=12, m=6, sig=params.sig, eps=params.eps).cut(rcut=params.rcut)

    phi1 = p.phi(params.r)
    dphi1 = -p.dphidr(params.r) / params.r

    np.testing.assert_allclose(phi0, phi1)
    np.testing.assert_allclose(dphi0, dphi1)


def test_nm_lfs(lj_lfs_params):
    params = lj_lfs_params

    phi0, dphi0 = params.get_phidphi()

    p = pots.Phi_nm(n=12, m=6, sig=params.sig, eps=params.eps).lfs(rcut=params.rcut)

    phi1 = p.phi(params.r)
    dphi1 = -p.dphidr(params.r) / params.r

    np.testing.assert_allclose(phi0, phi1)
    np.testing.assert_allclose(dphi0, dphi1)


def test_sw(sw_params):
    _do_test(sw_params, pots.Phi_sw, phidphi=False)


def test_yk(yk_params):
    _do_test(yk_params, pots.Phi_yk, kws=yk_params.asdict(), phidphi=False)


def test_hs(hs_params):
    _do_test(hs_params, pots.Phi_hs, phidphi=False)



from __future__ import annotations

import attrs
import numpy as np


def attrs_clear_cache(self, attribute, value):
    """
    clear out _cache if setting value
    """
    setattr(self, "_cache", {})
    return value


def optional_converter(converter):
    """
    Create a converter which can pass through None
    """

    def wrapped(value):
        if value is None or attrs.NOTHING:
            return value
        else:
            return converter(value)

    return wrapped


def field_formatter(fmt: str = "{:.5g}"):
    """Formatter for attrs field."""

    @optional_converter
    def wrapped(value):
        return fmt.format(value)

    return wrapped


def field_array_formatter(threshold=3, **kws):
    """Formatter for numpy array field"""

    @optional_converter
    def wrapped(value):
        with np.printoptions(threshold=threshold, **kws):
            return str(value)

    return wrapped


def private_field(init=False, repr=False, **kws):
    """
    Create a private attrs field.
    """
    return attrs.field(init=init, repr=repr, **kws)



from pathlib import Path

import numpy as np
import pandas as pd

# import analphipy.measures as measures
import analphipy.potentials as pots
from analphipy.norofrenkel import NoroFrenkelPair

data = Path(__file__).parent / "data"

nsamp = 20


def test_nf_sw():
    cols = ["sig", "eps", "lam", "B2"]
    df = pd.read_csv(data / "eff_sw_.csv").assign(beta=lambda x: 1.0 / x["temp"])

    if len(df) > nsamp:
        df = df.sample(nsamp)

    for (sig, eps, lam), g in df.groupby(["sig", "eps", "lam"]):

        p = pots.Phi_sw(sig=sig, eps=eps, lam=lam)

        a = NoroFrenkelPair(phi=p.phi, segments=p.segments, r_min=sig, phi_min=eps)

        out = a.table(g["beta"], cols)

        for col in cols:
            left = col + "_eff"
            np.testing.assert_allclose(g[left], out[col])


def test_nf_lj():

    cols = ["sig", "eps", "lam", "B2", "sig_dbeta"]

    df = (
        pd.read_csv(data / "eff_lj_.csv")
        .assign(beta=lambda x: 1.0 / x["temp"])
        .drop_duplicates()
        .assign(sig_dbeta_eff=lambda x: -x["dsig_eff"])
    )

    if len(df) > nsamp:
        df = df.sample(nsamp)
    sig = eps = 1.0
    p_base = pots.Phi_lj(sig=sig, eps=eps)

    for (rcut, tail), g in df.groupby(["rcut", "tail"]):

        if tail == "LFS":
            p = p_base.lfs(rcut=rcut)
        elif tail == "CUT":
            p = p_base.cut(rcut=rcut)
        elif tail == "LRC":
            p = p_base

        a = NoroFrenkelPair.from_phi(
            phi=p.phi, segments=p.segments, r_min=sig, bounds=[0.5, 1.5]
        )

        out = a.table(g["beta"], cols)

        for col in cols:
            left = col + "_eff"
            np.testing.assert_allclose(g[left], out[col], rtol=1e-3)


def test_nf_nm():

    cols = ["sig", "eps", "lam", "B2"]
    df = (
        pd.read_csv(data / "eff_lj_nm_.csv")
        .assign(beta=lambda x: 1.0 / x["temp"])
        .drop_duplicates()
    )

    if len(df) > nsamp:
        df = df.sample(nsamp)
    sig = eps = 1.0

    for (n_exp, m_exp), g in df.groupby(["n_exp", "m_exp"]):
        p = pots.Phi_nm(n=n_exp, m=m_exp, sig=sig, eps=eps)

        a = NoroFrenkelPair.from_phi(
            phi=p.phi, segments=p.segments, r_min=sig, bounds=[0.5, 1.5]
        )

        out = a.table(g["beta"], cols)

        for col in cols:
            left = col + "_eff"
            np.testing.assert_allclose(g[left], out[col], rtol=1e-3)


def test_nf_yk():
    cols = ["sig", "eps", "lam", "B2"]

    df = (
        pd.read_csv(data / "eff_yukawa_.csv")
        .assign(beta=lambda x: 1.0 / x["temp"])
        .drop_duplicates()
    )

    if len(df) > nsamp:
        df = df.sample(nsamp)
    sig = eps = 1.0

    for (z_yukawa), g in df.groupby(["z_yukawa"]):
        p = pots.Phi_yk(z=z_yukawa, sig=sig, eps=eps)

        a = NoroFrenkelPair(phi=p.phi, segments=p.segments, r_min=sig, phi_min=-1.0)

        out = a.table(g["beta"], cols)

        for col in cols:
            left = col + "_eff"
            np.testing.assert_allclose(g[left], out[col], rtol=1e-3)



# type: ignore
# from ._potentials import (
#     PhiBaseCuttable,
#     PhiBaseGenericCut,
#     PhiCut,
#     PhiLFS,
# )

from . import measures, norofrenkel
from .norofrenkel import NoroFrenkelPair
from .potentials import CubicTable, Phi_hs, Phi_lj, Phi_nm, Phi_sw, Phi_yk, factory_phi
from .potentials_base import PhiAnalytic, PhiBase, PhiCut, PhiGeneric, PhiLFS

# updated versioning scheme
try:
    from importlib.metadata import version as _version
except ImportError:
    # if the fallback library is missing, we are doomed.
    from importlib_metadata import version as _version  # type: ignore[no-redef]

try:
    __version__ = _version("analphipy")
except Exception:
    # Local copy or not installed with setuptools.
    # Disable minimum version checks on downstream libraries.
    __version__ = "999"


__all__ = [
    # "PhiBaseCuttable",
    # "PhiBaseGenericCut",
    "PhiCut",
    "PhiLFS",
    "PhiBase",
    "PhiAnalytic",
    "PhiGeneric",
    "Phi_lj",
    "Phi_nm",
    "Phi_sw",
    "Phi_hs",
    "Phi_yk",
    "CubicTable",
    "factory_phi",
    "NoroFrenkelPair",
    "measures",
    "norofrenkel",
    "__version__",
]



# type: ignore
# flake8: noqa

from __future__ import annotations

# import dataclasses
from collections.abc import Callable, Mapping
from dataclasses import dataclass
from functools import singledispatch

# from functools import partial
from typing import Optional, Sequence, Tuple, cast

import numpy as np

# from abc import abstractmethod
from custom_inherit import DocInheritMeta
from typing_extensions import Literal

from analphipy.norofrenkel import NoroFrenkelPair
from analphipy.utils import minimize_phi

from ._typing import Float_or_ArrayLike
from .cached_decorators import cached_clear, gcached
from .utils import (
    phi_to_phi_cut,
    phi_to_phi_lfs,
    segments_to_segments_cut,
    wca_decomp_att,
    wca_decomp_rep,
)

# * Utility function


class PhiBase:
    r"""
    Base class for working with pair potentials.

    This is the most general case, where you supply a callable function
    for ``phi`` and (optionally), ``dphi``.  In addition, extra parameters
    to handle the calculation of metrics is also included

    Parameters
    ----------
    phi : callable
        Function to calculate the pair potential.  This should have the form
        ``v = phi(r)``, where ``r`` can be a scalar or ndarray, and ``v`` is an ndarray.
    dphidr : callable, optional
        Function to calculate derivative of ``phi`` with respect to ``r``
        :math:`\frac{d \phi}{d r}`
    r_min : float, optional
        Position of minimum in ``phi``.
    phi_min : float, optional
        Value of ``phi`` at minimum.  If specified, but pass in value for ``r_min``, then
        ``phi_min = phi(r_min)``. This is available, to exactly set the value of
        the minimum.
    segments : sequence of floats, optional
        Bounds specifiying unique regions of the ``phi``.  Used in calculating various
        metric
    quad_kws : mapping, optional
        Extra arguments to :func:`analphipy.utils.quad_segments`

    meta : mapping, optional
        Optional mapping of any meta data to be passed along.
    """

    def __init__(
        self,
        phi: Callable,
        dphidr: Callable | None = None,
        r_min: float | None = None,
        phi_min: float | None = None,
        segments: Sequence[float] | None = None,
        quad_kws: Mapping | None = None,
        meta: Mapping | None = None,
    ):

        self._phi_func = phi
        self._dphidr_func = dphidr

        self._r_min = r_min
        self._phi_min = phi_min
        self._segments = segments
        self.meta = meta

        if quad_kws is None:
            quad_kws = {}
        else:
            quad_kws = dict(quad_kws)

        self.quad_kws = quad_kws

    @property
    def r_min(self):
        return self._r_min

    @r_min.setter
    @cached_clear()
    def r_min(self, val):
        self._r_min = val

    @property
    def phi_min(self):
        if self._phi_min is not None:
            return self._phi_min
        elif self._r_min is not None:
            return self.phi(self._r_min)
        else:
            return self._phi_min

    @phi_min.setter
    @cached_clear()
    def phi_min(self, val):
        self._phi_min = val

    @property
    def quad_kws(self):
        return self._quad_kws

    @quad_kws.setter
    @cached_clear()
    def quad_kws(self, val):
        if val is None:
            val = {}
        else:
            val = dict(val)

        self._quad_kws = val

    @property
    def segments(self):
        return self._segments

    @segments.setter
    @cached_clear()
    def segments(self, val):
        self._segments = val

    def phi(self, r: Float_or_ArrayLike) -> np.ndarray:
        r = np.asarray(r)
        return self._phi_func(r)

    def dphidr(self, r: Float_or_ArrayLike) -> np.ndarray:
        if self._dphidr_func is None:
            raise NotImplementedError("no method dphidr provided")
        else:
            r = np.asarray(r)
            return self._dphidr_func(r)

    def new_like(self, inherit_mins=True, **kws):
        """
        Create new object like ``self``.

        Any parameters specified will override parameters of ``self``.

        Parameters
        ----------
        inherit_mins : bool, default=True
            If True, inherit ``r_min`` and ``phi_min`` parameters unless explicitly specified.

        **kws :
            Parameters to explicitly set.
        """

        defaults = {
            "phi": self._phi_func,
            "dphidr": self._dphidr_func,
            "segments": self.segments,
            "quad_kws": self.quad_kws,
            "meta": self.meta,
        }

        if inherit_mins:
            defaults["r_min"], defaults["phi_min"] = self.r_min, self.phi_min

        kws = {**defaults, **kws}

        return type(self)(**kws)

    def cut(self, rcut):
        """
        Create cut versions of phi/dphidr
        """

        phi_cut, dphidr_cut, meta = phi_to_phi_cut(
            rcut=rcut, phi=self._phi_func, dphidr=self._dphidr_func, meta=self.meta
        )

        return self.new_like(
            phi=phi_cut,
            dphidr=dphidr_cut,
            segments=segments_to_segments_cut(self.segments, rcut),
            meta=meta,
            inherit_mins=False,
        )

    def lfs(self, rcut):
        """
        Create lfs version of phi/dphidr
        """

        if self._dphidr_func is None:
            raise ValueError("Must have ``dphidr`` function to apply lfs")

        phi_lfs, dphidr_lfs, meta = phi_to_phi_lfs(
            rcut=rcut, phi=self._phi_func, dphidr=self._dphidr_func, meta=self.meta
        )

        return self.new_like(
            phi=phi_lfs,
            dphidr=dphidr_lfs,
            segments=segments_to_segments_cut(self.segments, rcut),
            meta=meta,
            inherit_mins=False,
        )

    def minimize(
        self,
        r0: float | None = None,
        bounds: tuple[float, float] | Literal["None"] | None = None,
        **kws,
    ):
        """Determine position `r` where ``phi`` is minimized.

        Parameters
        ----------
        r0 : float
            Guess for position of minimum.
            Defaults to ``mean(bounds)``.
        bounds : tuple floats, {'None'}, optional
            If passed, should be of form ``bounds=(lower_bound, upper_bound)``.
            Defaults to ``(self.segments[0], self.segments[-1])``.
            If pass string value of `'None'`, then use ``bounds=[0, inf]``.
        **kws :
            Extra arguments to :func:`analphipy.minimize_phi`

        Returns
        -------
        rmin : float
            ``phi(rmin)`` is found location of minimum.
        phimin : float
            Value of ``phi(rmin)``, i.e., the value of ``phi`` at the minimum
        output :
            Output class from :func:`scipy.optimize.minimize`.


        See Also
        --------
        analphipy.minimize_phi
        """

        if bounds == "None":
            bounds = [0.0, np.inf]
        elif bounds is None:
            bounds = (self.segments[0], self.segments[-1])

        if r0 is None:
            r0 = np.mean(bounds)

        return minimize_phi(phi=self.phi, r0=r0, bounds=bounds, **kws)

    def set_min(
        self, r0: float | None = None, bounds: tuple[float, float] | None = None, **kws
    ):
        """
        Create new object with ``r_min`` and ``phi_min`` set by numerical minimization.

        See Also
        --------
        :meth:`minimize`
        """
        r_min, phi_min, output = self.minimize(r0=r0, bounds=bounds, **kws)
        return self.new_like(r_min=r_min, phi_min=phi_min)

    @gcached()
    def nf(self):
        """Noro frenkel object"""

        if self.r_min is None:
            raise ValueError("Must have r_min/phi_min specified")

        return NoroFrenkelPair(
            phi=self.phi,
            segments=self.segments,
            r_min=self.r_min,
            phi_min=self.phi_min,
            quad_kws=self.quad_kws,
        )

    @gcached()
    def measures(self):
        from .measures import Measures

        return Measures(self)

    @gcached()
    def wca_rep(self):
        """
        WCA repulsive
        """

        if self.r_min is None:
            raise ValueError("must specify r_min to access WCA decomp")

        phi, dphidr, meta = wca_decomp_rep(
            phi=self._phi_func,
            dphidr=self._dphidr_func,
            meta=self.meta,
            r_min=self.r_min,
            phi_min=self.phi_min,
        )

        return self.new_like(
            phi=phi,
            dphidr=dphidr,
            meta=meta,
            inherit_mins=False,
        )

    @gcached()
    def wca_att(self):
        """
        WCA repulsive
        """

        if self.r_min is None:
            raise ValueError("must specify r_min to access WCA decomp")

        phi, dphidr, meta = wca_decomp_att(
            phi=self._phi_func,
            dphidr=self._dphidr_func,
            meta=self.meta,
            r_min=self.r_min,
            phi_min=self.phi_min,
        )

        return self.new_like(
            phi=phi,
            dphidr=dphidr,
            meta=meta,
            inherit_mins=False,
        )

    def __repr__(self):

        if self.meta is None:
            return "<PhiBase>"
        else:
            kws = dict(self.meta)
            style = "_".join(kws.pop("style"))

            params = ", ".join([f"{k}={v}" for k, v in kws.items()])

            return f"<{style}: {params}>"


def factory_phi_lj(sig: float, eps: float, quad_kws: Mapping | None = None):
    r"""Lennard-Jones potential.

    .. math::

        \phi(r) = 4 \epsilon \left[ \left(\frac{{\sigma}}{{r}}\right)^{{12}} - \left(\frac{{\sigma}}{{r}}\right)^6\right]

    Parameters
    ----------
    sig : float
        Length parameter :math:`\sigma`.
    eps : float
        Energy parameter :math:`\epsilon`.
    """

    sigsq = sig * sig
    four_eps = 4.0 * eps

    def phi(r: Float_or_ArrayLike) -> np.ndarray:
        r = np.array(r)

        x2 = sigsq / (r * r)
        x6 = x2**3
        return cast(np.ndarray, four_eps * x6 * (x6 - 1.0))

    def dphidr(r: Float_or_ArrayLike) -> tuple[np.ndarray, np.ndarray]:
        """calculate phi and dphi (=-1/r dphi/dr) at particular r"""
        r = np.array(r)
        rinvsq = 1.0 / (r * r)

        x2 = sigsq * rinvsq
        x6 = x2 * x2 * x2

        dphidr = -12.0 * four_eps * x6 * (x6 - 0.5) / r
        return dphidr

    r_min = sig * 2.0 ** (1.0 / 6.0)

    phi_min = -eps

    segments = [0.0, np.inf]

    meta = {
        "style": ("lj",),
        "sig": sig,
        "eps": eps,
    }

    return PhiBase(
        phi=phi,
        dphidr=dphidr,
        r_min=r_min,
        phi_min=phi_min,
        segments=segments,
        quad_kws=quad_kws,
        meta=meta,
    )



import numpy as np
import pytest
from scipy.misc import derivative

import analphipy.potentials as pots
from analphipy.norofrenkel import NoroFrenkelPair


def _do_test(nf, beta, prop, dprop, dx=1e-8, rtol=1e-3):

    a = getattr(nf, dprop)(beta)

    f = getattr(nf, prop)

    b = derivative(f, beta, dx=dx)

    np.testing.assert_allclose(a, b, rtol=rtol)


def factory_lj(lfs=False, cut=False, **kws):

    if lfs or cut:
        rcut = kws.pop("rcut")
    else:
        rcut = None

    p = pots.Phi_lj(**kws)

    if cut:
        p = p.cut(rcut=rcut)
    if lfs:
        p = p.lfs(rcut=rcut)
    return p


BETAS = [0.1, 0.5, 1.0]


def test_nf_deriv_lj():

    sig = eps = 1.0

    p = pots.Phi_lj(sig=sig, eps=eps)

    nf = NoroFrenkelPair.from_phi(
        p.phi, p.segments, r_min=sig, bounds=[0.5 * sig, 1.5 * sig]
    )

    for prop in ["sig", "B2", "lam"]:
        for beta in BETAS:
            _do_test(nf, beta, prop, prop + "_dbeta", dx=1e-8, rtol=1e-2)


@pytest.mark.parametrize("rcut", [2.0, 3.0, 5.0])
def test_nf_deriv_lj_cut(rcut):

    sig = eps = 1.0

    p = pots.Phi_lj(sig=sig, eps=eps).cut(rcut)

    nf = NoroFrenkelPair.from_phi(
        p.phi, p.segments, r_min=sig, bounds=[0.5 * sig, 1.5 * sig]
    )

    for prop in ["sig", "B2", "lam"]:
        for beta in BETAS:
            _do_test(nf, beta, prop, prop + "_dbeta", dx=1e-8, rtol=1e-2)


@pytest.mark.parametrize("z", [1.0, 2.0, 4.0])
def test_nf_deriv_yk(z):

    sig = eps = 1.0

    p = pots.Phi_yk(sig=sig, eps=eps, z=z)

    nf = NoroFrenkelPair(p.phi, p.segments, r_min=sig, phi_min=-1.0)

    for prop in ["sig", "B2", "lam"]:
        for beta in BETAS:
            _do_test(nf, beta, prop, prop + "_dbeta", dx=1e-8, rtol=1e-2)



# type: ignore
from __future__ import annotations

from typing import Sequence, cast

import attrs
import numpy as np
from attrs import field
from custom_inherit import DocInheritMeta
from typing_extensions import Literal

from ._attrs_utils import field_formatter, optional_converter, private_field
from ._docstrings import docfiller_shared
from ._typing import Float_or_ArrayLike, Phi_Signature
from .measures import Measures
from .norofrenkel import NoroFrenkelPair
from .utils import minimize_phi


# * attrs utilities
@optional_converter
def segments_converter(segments):
    return tuple(float(x) for x in segments)


def validate_segments(self, attribute, segments):
    assert len(segments) >= 2, "Segments must be at least of length 2"


def _kw_only_field(kw_only=True, default=None, **kws):
    return attrs.field(kw_only=kw_only, default=default, **kws)


@attrs.frozen
@docfiller_shared
class _PhiBase(metaclass=DocInheritMeta(style="numpy_with_merge")):  # type: ignore
    """
    Base Class for working with potentials.

    Parameters
    ----------
    r_min : float, optional
        Position of minimum.
    phi_min : float, optional
        Value of ``phi`` at ``r_min``.
    {segments}
    """

    #: Position of minimum in :math:`\phi(r)`
    r_min: float | None = _kw_only_field(
        converter=optional_converter(float), repr=field_formatter()
    )

    #: Value of ``phi`` at minimum.
    phi_min: float | None = _kw_only_field(
        converter=optional_converter(float), repr=field_formatter()
    )

    #: Integration limits
    segments: Sequence[float] | None = _kw_only_field(
        converter=segments_converter,
    )

    def asdict(self) -> dict:
        """Convert object to dictionary"""
        return attrs.asdict(self, filter=self._get_smart_filter())

    def new_like(self, **kws):
        """Create a new object with optional parameters

        Parameters
        ----------
        **kws
            `attribute`, `value` pairs.
        """
        return attrs.evolve(self, **kws)

    def assign(self, **kws):
        """Alias to :meth:`new_like`"""
        return self.new_like(**kws)

    def _immutable_setattrs(self, **kws):
        """
        Set attributes of frozen attrs class.

        This should only be used to set 'derived' attributes on creation.

        Parameters
        ----------
        **kws :
            `key`, `value` pairs to be set.  `key` should be the name of the attribute to be set,
            and `value` is the value to set `key` to.

        Examples
        --------
        >>> @attrs.frozen
        ... class Derived(PhiBase):
        ...     _derived = attrs.field(init=False)
        ...
        ...     def __attrs_post_init__(self):
        ...         self._immutable_setattrs(_derived=self.r_min + 10)

        >>> x = Derived(r_min=5)
        >>> x._derived
        15.0
        """

        for key, value in kws.items():
            object.__setattr__(self, key, value)

    def _get_smart_filter(
        self, include=None, exclude=None, exclude_private=True, exclude_no_init=True
    ):
        """
        create a filter to include exclude names

        Parameters
        ----------
        include : sequence of str
            Names to include
        exclude : sequence of str
            names to exclude
        exclude_private : bool, default=True
            If True, exclude any names starting with '_'
        exclude_no_init : bool, default=True
            If True, exclude any fields defined with ``field(init=False)``.

        Notes
        -----
        Precidence is in order
        `include, exclude, exclude_private exclude_no_init`.
        That is, if a name is in include and exclude and is private/no_init,
        it will be included
        """
        fields = attrs.fields(type(self))

        if include is None:
            include = []
        elif isinstance(include, str):
            include = [include]

        if exclude is None:
            exclude = []
        elif isinstance(exclude, str):
            exclude = [exclude]

        includes = []
        for f in fields:

            if f.name in include:
                includes.append(f)

            elif f.name in exclude:
                pass

            elif exclude_private and f.name.startswith("_"):
                pass

            elif exclude_no_init and not f.init:
                pass

            else:
                includes.append(f)
        return attrs.filters.include(*includes)

    def phi(self, r: Float_or_ArrayLike) -> np.ndarray:
        """
        Pair potential

        Parameters
        ----------
        r : array-like
            pair separation

        Returns
        -------
        phi : ndarray
            Evaluated pair potential.
        """

        raise NotImplementedError("Must implement in subclass")

    def dphidr(self, r: Float_or_ArrayLike) -> np.ndarray:
        r"""
        Derivative of pair potential.

        This returns the value of :math:`d \phi(r) / dr`



        Parameters
        ----------
        r : array-like
            pair separation

        Returns
        -------
        dphidr : ndarray
            Pair potential values.

        """
        raise NotImplementedError("Must implement in subclass")

    def minimize(
        self,
        r0: float | Literal["mean"],
        bounds: tuple[float, float] | Literal["segments"] | None = None,
        **kws,
    ):
        """Determine position `r` where ``phi`` is minimized.

        Parameters
        ----------
        r0 : float
            Guess for position of minimum.
            If value is `'mean'`, then use ``r0 = mean(bounds)``.

        bounds : tuple floats, {'segments'}, optional
            Bounds for minimization search.
            If tuple, should be of form ``bounds=(lower_bound, upper_bound)``.
            If `'segments`, then ``bounds=(segments[0], segments[-1])``.
            If None, no bounds used.
        **kws :
            Extra arguments to :func:`analphipy.minimize_phi`

        Returns
        -------
        rmin : float
            ``phi(rmin)`` is found location of minimum.
        phimin : float
            Value of ``phi(rmin)``, i.e., the value of ``phi`` at the minimum
        output :
            Output class from :func:`scipy.optimize.minimize`.


        See Also
        --------
        ~analphipy.utils.minimize_phi
        """

        if bounds == "segments":
            if self.segments is not None:
                bounds = (self.segments[0], self.segments[-1])
            else:
                bounds = None

        if r0 == "mean":
            if bounds is not None:
                r0 = np.mean(bounds)
            else:
                r0 = None

        return minimize_phi(self.phi, r0=r0, bounds=bounds, **kws)

    def assign_min_numeric(
        self, r0: float, bounds: tuple[float, float] | None = None, **kws
    ):
        """
        Create new object with minima set by numerical minimization

        call :meth:`minimize`
        """

        r_min, phi_min, output = self.minimize(r0=r0, bounds=bounds, **kws)
        return self.new_like(r_min=r_min, phi_min=phi_min)

    def to_nf(self, **kws):
        """
        Create a :class:`analphipy.NoroFrenkelPair` object.

        Parameters
        ---------
        **kws :
            Extra arguments to :class:`analphipy.NoroFrenkelPair` constructor.
            parameters `phi`, `semgnets`, `r_min', `phi_min` default to values from `self`.

        Returns
        -------
        nf : :class:`analphipy.NoroFrenkelPair`

        """
        if self.r_min is None:
            raise ValueError("must set `self.r_min` to use NoroFrenkel")

        for k in ["phi", "segments", "r_min", "phi_min"]:
            if k not in kws:
                kws[k] = getattr(self, k)

        return NoroFrenkelPair(**kws)

    def to_measures(self, **kws):
        """
        Create a :class:`analphipy.measures.Measures` object.

        Parameters
        ----------
        **kws :
            Extra arguments to :class:`analphipy.Measures` constructor.
            parameters `phi`, `semgnets` default to values from `self`.

        Returns
        -------
        nf : :class:`analphipy.measures.Measures`
        """

        for k in ["phi", "segments"]:
            if k not in kws:
                kws[k] = getattr(self, k)

        return Measures(**kws)


@attrs.frozen
class _PhiBaseCut(_PhiBase):
    """
    Base Class for creating cut potential from base potential


    phi_cut(r) = phi(r) + _vcorrect(r) if r <= rcut
               = 0 otherwise


    dphi_cut(r) = dphi(r) + _dvcorrect(r) if r <= rcut
                = 0 otherwise

    So, for example, for just cut, `_vcorrect(r) = -v(rcut)`, `dvrcut=0`,
    and for lfs, `_vcorrect(r) = -v(rcut) - dv(rcut)/dr (r - rcut)`
    `_dvcorrect(r) = ...`


    Parameters
    ----------
    phi_base : PhiBase
        Instance of (sub)class of :class:`analphipy.PhiBase`.
    rcut : float
        Position to cut the potential.


    """

    #: Instance of (sub)class of :class:`analphipy.PhiBase`
    phi_base: PhiBase
    #: Position to cut the potential
    rcut: float = field(converter=float)
    #: Integration limits
    segments: float = private_field()

    def __attrs_post_init__(self):
        self._immutable_setattrs(
            segments=tuple(x for x in self.phi_base.segments if x < self.rcut)
            + (self.rcut,)
        )

    def phi(self, r: Float_or_ArrayLike) -> np.ndarray:
        r = np.asarray(r)
        v = np.empty_like(r)

        left = r <= self.rcut
        right = ~left

        v[right] = 0.0

        if np.any(left):
            v[left] = self.phi_base.phi(r[left]) + self._vcorrect(r[left])
        return v

    def dphidr(self, r: Float_or_ArrayLike) -> np.ndarray:
        r = np.asarray(r)
        dvdr = np.empty_like(r)

        left = r <= self.rcut
        right = ~left

        dvdr[right] = 0.0
        dvdr[left] = self.phi_base.dphidr(r[left]) + self._dvdrcorrect(r[left])

        return dvdr

    def _vcorrect(self, r):
        raise NotImplementedError

    def _dvdrcorrect(self, r):
        raise NotImplementedError


@attrs.frozen
class PhiCut(_PhiBaseCut):
    r"""
    Pair potential cut at position ``r_cut``.

    .. math::

        \phi_{\rm cut}(r) =
            \begin{cases}
                \phi(r) - \phi(r_{\rm cut})  & r < r_{\rm cut} \\
                0 & r_{\rm cut} \leq r
            \end{cases}

    Parameters
    ----------
    phi_base : PhiBaseCuttable instance
        Potential class to perform cut on.
    rcut : float
        Where to 'cut' the potential.
    """

    _vcut: float = private_field()

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._immutable_setattrs(_vcut=self.phi_base.phi(self.rcut))

    def _vcorrect(self, r: np.ndarray) -> np.ndarray:
        return -cast(np.ndarray, self._vcut)

    def _dvdrcorrect(self, r: np.ndarray) -> np.ndarray:
        return np.array(0.0)


@attrs.frozen
class PhiLFS(_PhiBaseCut):
    r"""
    Pair potential cut and linear force shifted at ``r_cut``.

    .. math::

        \phi_{\rm lfs}(r) =
            \begin{cases}
                \phi(r)
                - \left( \frac{d \phi}{d r} \right)_{\rm cut} (r - r_{\rm cut})
                - \phi(r_{\rm cut}) & r < r_{\rm cut} \\
                0 & r_{\rm cut} < r
            \end{cases}
    """

    _vcut: float = private_field()
    _dvdrcut: float = private_field()

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._immutable_setattrs(
            _vcut=self.phi_base.phi(self.rcut), _dvdrcut=self.phi_base.dphidr(self.rcut)
        )

    def _vcorrect(self, r: np.ndarray) -> np.ndarray:
        out = -(self._vcut + self._dvdrcut * (r - self.rcut))
        return cast(np.ndarray, out)

    def _dvdrcorrect(self, r: np.ndarray) -> np.ndarray:
        out = -self._dvdrcut
        return cast(np.ndarray, out)


@attrs.frozen
class PhiBase(_PhiBase):
    def cut(self, rcut):
        """Create a :class:`PhiCut` potential."""
        return PhiCut(phi_base=self, rcut=rcut)

    def lfs(self, rcut):
        """Create a :class:`PhiLFS` potential."""
        return PhiLFS(phi_base=self, rcut=rcut)


@attrs.frozen
class PhiGeneric(PhiBase):
    """
    Class to define potential using callables.

    Parameters
    ----------
    phi_func : Callable
        Function ``phi(r)``
    dphidr : Callable, optional
        Optional function ``dphidr(r)``.
    """

    #: Function :math:`\phi(r)`
    phi_func: Phi_Signature

    #: Function :math:`d\phi(r)/dr`
    dphidr_func: Phi_Signature | None = None

    def phi(self, r: Float_or_ArrayLike) -> np.ndarray:
        r = np.asarray(r)
        return self.phi_func(r)

    def dphidr(self, r: Float_or_ArrayLike) -> np.ndarray:
        if self.dphidr_func is None:
            raise ValueError("Must specify dphidr_func")
        r = np.asarray(r)
        return self.dphidr_func(r)


@attrs.frozen
class PhiAnalytic(PhiBase):
    """
    Base class for defining analytic potentials.

    Notes
    -----
    Specific subclasses should set values for ``r_min``,
    ``phi_min``, and ``segments``, as well as
    forms for ``phi`` and ``dphidr``.

    """

    #: Position of minimum in :math:`\phi(r)`
    r_min: float = field(init=False, repr=field_formatter())
    #: Value of ``phi`` at minimum.
    phi_min: float = field(init=False, repr=False)
    #: Integration limits.
    segments: float = field(init=False)



from pathlib import Path

import numpy as np
import pandas as pd

import analphipy.measures as measures
import analphipy.potentials as pots
from analphipy.norofrenkel import NoroFrenkelPair

data = Path(__file__).parent / "data"

nsamp = 20


def test_B2_sw():

    df = (
        pd.read_csv(data / "vir_sw_.csv")
        .assign(beta=lambda x: 1.0 / x["temp"])
        .drop_duplicates()
    )

    if len(df) > nsamp:
        df = df.sample(nsamp)
    for (sig, eps, lam), g in df.groupby(["sig", "eps", "lam"]):

        p = pots.Phi_sw(sig=sig, eps=eps, lam=lam)

        a = NoroFrenkelPair(phi=p.phi, segments=p.segments, r_min=sig, phi_min=eps)

        out = a.table(g["beta"], ["B2", "B2_dbeta"])

        np.testing.assert_allclose(g["Bn_2"], out["B2"])
        np.testing.assert_allclose(-g["dBn_2"], out["B2_dbeta"])


def test_B2_lj():

    df = (
        pd.read_csv(data / "vir_lj_.csv")
        .assign(beta=lambda x: 1.0 / x["temp"])
        .drop_duplicates()
        .sample(40)
    )

    if len(df) > nsamp:
        df = df.sample(nsamp)
    sig = eps = 1.0
    p_base = pots.Phi_lj(sig=sig, eps=eps)

    for (rcut, tail), g in df.groupby(["rcut", "tail"]):

        if tail == "LFS":
            p = p_base.lfs(rcut=rcut)
        elif tail == "CUT":
            p = p_base.cut(rcut=rcut)
        elif tail == "LRC":
            p = p_base

        a = NoroFrenkelPair.from_phi(
            phi=p.phi, segments=p.segments, r_min=sig, bounds=[0.5, 1.5]
        )

        out = a.table(g["beta"], ["B2", "B2_dbeta"])

        np.testing.assert_allclose(g["Bn_2"], out["B2"], rtol=1e-3)
        np.testing.assert_allclose(-g["dBn_2"], out["B2_dbeta"], rtol=1e-3)


def test_B2_nm():

    df = (
        pd.read_csv(data / "vir_lj_nm_.csv")
        .assign(beta=lambda x: 1.0 / x["temp"])
        .drop_duplicates()
    )

    if len(df) > nsamp:
        df = df.sample(nsamp)
    sig = eps = 1.0

    for (n_exp, m_exp), g in df.groupby(["n_exp", "m_exp"]):
        p = pots.Phi_nm(n=n_exp, m=m_exp, sig=sig, eps=eps)

        a = NoroFrenkelPair.from_phi(
            phi=p.phi, segments=p.segments, r_min=sig, bounds=[0.5, 1.5]
        )

        out = a.table(g["beta"], ["B2", "B2_dbeta"])

        np.testing.assert_allclose(g["Bn_2"], out["B2"], rtol=1e-3)
        np.testing.assert_allclose(-g["dBn_2"], out["B2_dbeta"], rtol=1e-3)


def test_B2_yk():
    df = (
        pd.read_csv(data / "vir_yukawa_.csv")
        .assign(beta=lambda x: 1.0 / x["temp"])
        .drop_duplicates()
    )

    if len(df) > nsamp:
        df = df.sample(nsamp)
    sig = eps = 1.0

    for (z_yukawa), g in df.groupby(["z_yukawa"]):
        p = pots.Phi_yk(z=z_yukawa, sig=sig, eps=eps)

        a = NoroFrenkelPair(phi=p.phi, segments=p.segments, r_min=sig, phi_min=-1.0)

        out = a.table(g["beta"], ["B2", "B2_dbeta"])

        np.testing.assert_allclose(g["Bn_2"], out["B2"], rtol=1e-6)
        np.testing.assert_allclose(-g["dBn_2"], out["B2_dbeta"], rtol=1e-6)


def test_B2_hs():

    sig = np.random.rand()

    p = pots.Phi_hs(sig=sig)

    B2a = 2 * np.pi / 3.0 * sig**3

    B2b = measures.secondvirial(p.phi, beta=1.0, segments=[0.0, sig])

    np.testing.assert_allclose(B2a, B2b)

    B2_dbeta = measures.secondvirial_dbeta(p.phi, beta=1.0, segments=[0.0, sig])

    np.testing.assert_allclose(0.0, B2_dbeta)



"""
routines to define a cached class without needing to subclass Cached class
"""
from __future__ import absolute_import

from functools import wraps
from inspect import signature

__all__ = ["cached", "cached_func", "cached_clear", "gcached"]


def gcached(key=None, prop=True):
    def wrapper(func):
        if prop:
            wrapped = property(cached(key)(func))
        else:
            wrapped = cached_func(key)(func)
        return wrapped

    return wrapper


def gcached_use_cache(key=None, prop=True):
    """
    same as gcached, but only cache func/prop if
    has attribute self._use_cache and self._use_cache is True
    """

    def wrapper(func):
        if prop:
            wrapped = property(cached(key, check_use_cache=True)(func))
        else:
            wrapped = cached_func(key, check_use_cache=True)(func)
        return wrapped

    return wrapper


def cached(key=None, check_use_cache=False):
    """Decorator to cache a property within a class

    Requires the Class to have a cache dict called ``_cache``.

    Notes
    -----
    Usage::

        class A(object):
           def__init__(self):
               # this isn't strictly needed as it will be created on demand
               self._cache = dict()

           @property
           @cached('keyname')
           def size(self):
               # This code gets ran only if the lookup of keyname fails
               # After this code has been ran once, the result is stored in
               # _cache with the key: 'keyname'
               size = 10.0

            #no aguments implies give cache function name
            @property
            @cached()
            def myprop(self):
                #results in _cache['myprop']

    See Also
    --------
    cached_clear : corresponding decorator to clear cache
    cached_func : decorator for cache creation of function
    """

    def cached_lookup(func):
        if key is None:
            _key = func.__name__
        else:
            _key = key

        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if (not check_use_cache) or (getattr(self, "_use_cache", False)):
                try:
                    return self._cache[_key]
                except AttributeError:
                    self._cache = dict()
                except KeyError:
                    pass

                self._cache[_key] = ret = func(self, *args, **kwargs)
                return ret
            else:
                return func(self, *args, **kwargs)

        return wrapper

    return cached_lookup


def cached_func(key=None, check_use_cache=False):
    """Decorator to cache a function within a class

    Requires the Class to have a cache dict called ``_cache``.

    Notes
    -----
    Usage::

        class A(object):
           def __init__(self):
              pass

           @cached_func('keyname')
           def amethod(self, *args):
               # This code gets ran only if the lookup of keyname fails
               # After this code has been ran once, the result is stored in
               # _cache with the key: 'keyname'
               # a long calculation...
               return long_calc(self,val)
               # if already executed result in _cache[('keyname',) + args]

            #no aguments implies give cache function name
            @property
            @cached()
            def myprop(self):
                #results in _cache['myprop']


    See also
    --------
    cached_clear : corresponding decorator to remove cache

    cached : decorator for properties

    """

    def cached_lookup(func):
        if key is None:
            _key = func.__name__
        else:
            _key = key

        # use signature
        # see james powell
        # https://youtu.be/gzbmleam8gs?t=7596
        # 2:09:00
        bind = signature(func).bind

        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if (not check_use_cache) or (getattr(self, "_use_cache", False)):

                params = bind(self, *args, **kwargs)
                params.apply_defaults()
                key_func = (_key, params.args[1:], frozenset(params.kwargs.items()))

                try:
                    return self._cache[key_func]
                except TypeError:
                    # this means that key_func is bad hash
                    return func(self, *args, **kwargs)
                except AttributeError:
                    self._cache = dict()
                except KeyError:
                    pass

                self._cache[key_func] = ret = func(self, *args, **kwargs)
                return ret
            else:
                return func(self, *args, **kwargs)

        return wrapper

    return cached_lookup


def cached_clear(*keys):
    """
    Decorator to clear self._cache of specified properties

    Parameters
    ----------
    *keys : arguments
        remove these keys from cache.  if len(keys)==0, remove all keys.


    Examples
    --------
    Usage::

        class tmp(object)
            ...

            @property
            @cached()
            def a(self):
                ...

            @x.setter
            @cached_clear('a','b')
            def x(self,val):
                #....
                #deletes self._cache['a'],self,_cache['b']
                #if args are empty, remove all


    See Also
    --------
    cached : corresponding decorator for cache creation of property
    cached_func : decorator for cache creation of function
    """

    def cached_clear(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            # self._clear_caches(*keys)
            # clear out keys
            if len(keys) == 0 or not hasattr(self, "_cache"):
                self._cache = dict()
            else:
                for name in keys:
                    try:
                        del self._cache[name]
                    except KeyError:
                        pass

                # functions
                keys_tuples = [
                    keys for k in self._cache if isinstance(k, tuple) and k[0] in keys
                ]
                for name in keys_tuples:
                    del self._cache[name]
            return func(self, *args, **kwargs)

        return wrapper

    return cached_clear
